/* vi: set sw=4 ts=4: */
/* serverpacket.c
 *
 * Construct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "common.h"
#include "dhcpd.h"
#include "options.h"

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>

#if HAS_DEVICE_MANAGER
extern char* brtable_find_if_by_mac(unsigned char *macAddr);
#endif

#if HAS_USER_MANUAL_BOOTPFLAG
enum {
	BOOTP_FLAG_AUTO  = 0,
	BOOTP_FLAG_BCAST = 1,
	BOOTP_FLAG_UCAST = 2,
};
#endif

/*****************************************************************
* NAME: SystemLog
* ---------------------------------------------------------------
* FUNCTION:  
* INPUT:    
* OUTPUT:   
* Author:   
* Modify:   
****************************************************************/
int SystemLog(char *format, ...)
{
    va_list ap;
    char SystemLogBuf[256];

    va_start(ap, format);
    vsnprintf(SystemLogBuf, sizeof(SystemLogBuf), format, ap);
    /* printf("----->%s\n", SystemLogBuf); */
    openlog( "systemd", LOG_NDELAY, LOG_LOCAL1);

    syslog(LOG_INFO, SystemLogBuf);
    closelog();

    return 0;
}

/* send a packet to giaddr using the kernel ip stack */
static int send_packet_to_relay(struct dhcpMessage *payload)
{
	DEBUG("Forwarding packet to relay");

	return udhcp_kernel_packet(payload, server_config.server, SERVER_PORT,
			payload->giaddr, SERVER_PORT);
}


/* send a packet to a specific arp address and ip address by creating our own ip packet */
static int send_packet_to_client(struct dhcpMessage *payload, int force_broadcast)
{
	const uint8_t *chaddr;
	uint32_t ciaddr;

	if (force_broadcast) {
		DEBUG("broadcasting packet to client (NAK)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr) {
		DEBUG("unicasting packet to client ciaddr");
		ciaddr = payload->ciaddr;
		chaddr = payload->chaddr;
	} else if (ntohs(payload->flags) & BROADCAST_FLAG) {
		DEBUG("broadcasting packet to client (requested)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else {
		DEBUG("unicasting packet to client yiaddr");
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
	}
	return udhcp_raw_packet(payload, server_config.server, SERVER_PORT,
			ciaddr, CLIENT_PORT, chaddr, server_config.ifindex);
}


/* send a dhcp packet, if force broadcast is set, the packet will be broadcast to the client */
static int send_packet(struct dhcpMessage *payload, int force_broadcast)
{
	int ret;

	if (payload->giaddr)
		ret = send_packet_to_relay(payload);
	else ret = send_packet_to_client(payload, force_broadcast);
	return ret;
}


static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type)
{
	udhcp_init_header(packet, type);
	packet->xid = oldpacket->xid;
	memcpy(packet->chaddr, oldpacket->chaddr, 16);

#if HAS_USER_MANUAL_BOOTPFLAG
	if(server_config.bootp_type == BOOTP_FLAG_BCAST) 
	{
		packet->flags = htons(0x8000);
	}
	else if(server_config.bootp_type == BOOTP_FLAG_UCAST) 
	{
		packet->flags = htons(0x0000);
	}
	else
	{
		packet->flags = oldpacket->flags;
	}
#else
	packet->flags = oldpacket->flags;
#endif

	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config.server);
}


/* add in the bootp options */
static void add_bootp_options(struct dhcpMessage *packet)
{
	packet->siaddr = server_config.siaddr;
	if (server_config.sname)
		strncpy((char*)packet->sname, server_config.sname, sizeof(packet->sname) - 1);
	if (server_config.boot_file)
		strncpy((char*)packet->file, server_config.boot_file, sizeof(packet->file) - 1);
}


/* send a DHCP OFFER to a DHCP DISCOVER */
int sendOffer(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct dhcpOfferedAddr *lease = NULL;
	uint32_t req_align, lease_time_align = server_config.lease;
	uint8_t *req, *lease_time;
	struct option_set *curr;
	struct in_addr addr;
	uint8_t *hostname, host[20], len;

	uint32_t static_lease_ip;

	init_packet(&packet, oldpacket, DHCPOFFER);

	static_lease_ip = getIpByMac(server_config.static_leases, oldpacket->chaddr);

	/* ADDME: if static, short circuit */
	if (!static_lease_ip) {
		/* the client is in our lease/offered table */
		lease = find_lease_by_chaddr(oldpacket->chaddr);

		if (lease && !reservedIp(server_config.static_leases, lease->yiaddr)) {
			if (!lease_expired(lease))
#if 1 //cfho 2008-0109
                lease_time_align = lease->expires - monotonic_sec();
#else
				lease_time_align = lease->expires - time(0);
#endif
			packet.yiaddr = lease->yiaddr;
		/* Or the client has a requested ip */
		} else if ((req = get_option(oldpacket, DHCP_REQUESTED_IP))
		 /* Don't look here (ugly hackish thing to do) */
		 && memcpy(&req_align, req, 4)
		 /* and the ip is in the lease range */
		 && ntohl(req_align) >= server_config.start_ip
		 && ntohl(req_align) <= server_config.end_ip
		 && !reservedIp(server_config.static_leases, req_align) /* Check that its not a static lease */
		 /* and is not already taken/offered */
		 && (!(lease = find_lease_by_yiaddr(req_align))
			/* or its taken, but expired */ /* ADDME: or maybe in here */
			|| lease_expired(lease))
		) {
		    if(req_align==server_config.server)
            {
                packet.yiaddr = find_address(1); /* reserved for DHCP Server, find another one */
            }else
            {
			    if( 0==arpping(req_align, server_config.server, 
				    	server_config.arp, server_config.interface) )
			    {
				    /* get Arp, This IP was used, find another one */
				    packet.yiaddr = find_address(1);
			    }else
			    {
				    packet.yiaddr = req_align; /* FIXME: oh my, is there a host using this IP? */
			    }
			}
			/* otherwise, find a free IP */
		} else {
			/* Is it a static lease? (No, because find_address skips static lease) */
			packet.yiaddr = find_address(0);
			/* try for an expired lease */
			if (!packet.yiaddr)
				packet.yiaddr = find_address(1);
		}

		if (!packet.yiaddr) {
			bb_error_msg("no IP addresses to give - OFFER abandoned");
			return -1;
		}
		
		hostname  = get_option(oldpacket, DHCP_HOST_NAME);

        if(hostname)
        {
            /* Got DHCP Host Name Option */
            len = *(hostname-1);
            if(len<20)
            {
                memset(host, 0, 20);
                memcpy(host, hostname, len);
                hostname = &host[0];
            }else
            {
                hostname = NULL;
            }
        }
		if (!add_lease(packet.chaddr, packet.yiaddr, server_config.offer_time, hostname)) {
			bb_error_msg("lease pool is full - OFFER abandoned");
			return -1;
		}
		lease_time = get_option(oldpacket, DHCP_LEASE_TIME);
		if (lease_time) {
			memcpy(&lease_time_align, lease_time, 4);
			lease_time_align = ntohl(lease_time_align);
			if (lease_time_align > server_config.lease)
				lease_time_align = server_config.lease;
		}

		/* Make sure we aren't just using the lease time from the previous offer */
		if (lease_time_align < server_config.min_lease)
			lease_time_align = server_config.lease;
		/* ADDME: end of short circuit */
	} else {
		/* It is a static lease... use it */
		packet.yiaddr = static_lease_ip;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
	bb_info_msg("Sending OFFER of %s", inet_ntoa(addr));
        /* systemlog message */
        SystemLog("DHCP Server, Sending OFFER of %s", inet_ntoa(addr));

	return send_packet(&packet, 0);
}


int sendNAK(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;

	init_packet(&packet, oldpacket, DHCPNAK);

	DEBUG("Sending NAK");
	return send_packet(&packet, 1);
}


int sendACK(struct dhcpMessage *oldpacket, uint32_t yiaddr, char * hostname)
{
	struct dhcpMessage packet;
	struct option_set *curr;
	uint8_t *lease_time;
	uint32_t lease_time_align = server_config.lease;
	struct in_addr addr;
	#if HAS_DEVICE_MANAGER
	uint32_t static_lease_ip;
	struct dhcpOfferedAddr *lease, static_lease;
	uint8_t buf[50],cmd_buf[100];
	#endif

	init_packet(&packet, oldpacket, DHCPACK);
	packet.yiaddr = yiaddr;

	lease_time = get_option(oldpacket, DHCP_LEASE_TIME);
	if (lease_time) {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
		else if (lease_time_align < server_config.min_lease)
			lease_time_align = server_config.lease;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
	bb_info_msg("Sending ACK to %s", inet_ntoa(addr));
        /* systemlog message */
        SystemLog("DHCP Server, Sending ACK of %s", inet_ntoa(addr));

#if DHCPD_REQUEST_INIC_CLIENTINFO
	// 2011-07-04 Norkay, Remove request clientinfo to udhcpd from https.
	system("sysconf_iNIC request clientinfo &");
#endif
        
	if (send_packet(&packet, 0) < 0)
		return -1;

#if HAS_DEVICE_MANAGER
	/* Look for a static lease */
	static_lease_ip = getIpByMac(server_config.static_leases, &packet.chaddr);

	if (static_lease_ip)
	{
		lease = &static_lease;
	} 
	else 
	{
		lease = find_lease_by_chaddr(packet.chaddr);
	}
	
	if(lease)
		lease->expires_notify_dm = 0;

	sprintf(cmd_buf,"dm_cli inform dhcp_ack ");

	//mac
	sprintf(buf,"\"%02X:%02X:%02X:%02X:%02X:%02X&", packet.chaddr[0], packet.chaddr[1], packet.chaddr[2], 
				packet.chaddr[3], packet.chaddr[4], packet.chaddr[5]);
	strcat(cmd_buf,buf);

	//ip
	addr.s_addr = packet.yiaddr;
	sprintf(buf,"%s&", inet_ntoa(addr));
	strcat(cmd_buf,buf);
	
	//hostname , if hostname is empty , set "---" as hostname
	if(0==strlen(hostname))
	    strcat(cmd_buf,"---&");
	else
	{
	    sprintf(buf,"%s&", hostname);
	    strcat(cmd_buf,buf);
	}

	//interface name	
	sprintf(buf," %s&\" ", brtable_find_if_by_mac(&(packet.chaddr[0])));
	strcat(cmd_buf,buf);

	bb_info_msg("%s from dhcp\n",cmd_buf);

	system(cmd_buf);
#endif

	add_lease(packet.chaddr, packet.yiaddr, lease_time_align, hostname);
	if (ENABLE_FEATURE_UDHCPD_WRITE_LEASES_EARLY) {
		/* rewrite the file with leases at every new acceptance */
		write_leases();
	}

	return 0;
}


int send_inform(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct option_set *curr;

	init_packet(&packet, oldpacket, DHCPACK);

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	return send_packet(&packet, 0);
}
